#
#  Copyright © 2010—2023 Andrey Mikhaylenko and contributors
#
#  This file is part of Argh.
#
#  Argh is free software under terms of the GNU Lesser
#  General Public License version 3 (LGPLv3) as published by the Free
#  Software Foundation. See the file README.rst for copying conditions.
#
"""
Dispatching
~~~~~~~~~~~
"""

import argparse
import inspect
import io
import sys
import warnings
from types import GeneratorType
from typing import IO, Any, Callable, Dict, Iterator, List, Optional, Tuple

from argh.assembling import NameMappingPolicy, add_commands, set_default_command
from argh.completion import autocomplete
from argh.constants import (
    ATTR_WRAPPED_EXCEPTIONS,
    ATTR_WRAPPED_EXCEPTIONS_PROCESSOR,
    DEST_FUNCTION,
    PARSER_FORMATTER,
)
from argh.exceptions import CommandError, DispatchingError

__all__ = [
    "ArghNamespace",
    "dispatch",
    "dispatch_command",
    "dispatch_commands",
    "parse_and_resolve",
    "run_endpoint_function",
    "PARSER_FORMATTER",
    "EntryPoint",
]


class ArghNamespace(argparse.Namespace):
    """
    A namespace object which collects the stack of functions.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._functions_stack: List[Callable] = []

    def __setattr__(self, key: str, value: Any) -> None:
        if key == DEST_FUNCTION:
            # don't register the function under DEST_FUNCTION name.
            # If `ArgumentParser.parse_known_args()` sees that we already have
            # such attribute, it skips it.  However, it goes from the topmost
            # parser to subparsers.  We need the function mapped to the
            # subparser.  So we fool the `ArgumentParser` and pretend that we
            # didn't get a DEST_FUNCTION attribute; however, in fact we collect
            # all its values in a stack.  The last item in the stack would be
            # the function mapped to the innermost parser — the one we need.
            self._functions_stack.append(value)
        else:
            super().__setattr__(key, value)

    def get_function(self) -> Callable:
        return self._functions_stack[-1]


def dispatch(
    parser: argparse.ArgumentParser,
    argv: Optional[List[str]] = None,
    add_help_command: bool = False,
    completion: bool = True,
    output_file: IO = sys.stdout,
    errors_file: IO = sys.stderr,
    raw_output: bool = False,
    namespace: Optional[argparse.Namespace] = None,
    skip_unknown_args: bool = False,
    always_flush: bool = False,
) -> Optional[str]:
    """
    Parses given list of arguments using given parser, calls the relevant
    function and prints the result.

    Internally calls :func:`~argh.dispatching.parse_and_resolve` and then
    :func:`~argh.dispatching.run_endpoint_function`.

    :param parser:

        the ArgumentParser instance.

    :param argv:

        a list of strings representing the arguments. If `None`, ``sys.argv``
        is used instead. Default is `None`.

    :param add_help_command:

        if `True`, converts first positional argument "help" to a keyword
        argument so that ``help foo`` becomes ``foo --help`` and displays usage
        information for "foo". Default is `False`.

        .. versionchanged:: 0.30

           The default value is now ``False`` instead of ``True``.

        .. deprecated:: 0.30

           This argument will be removed in v.0.31.  The user is expected to
           use ``--help`` instead of ``help``.

    :param output_file:

        A file-like object for output. If `None`, the resulting lines are
        collected and returned as a string. Default is ``sys.stdout``.

    :param errors_file:

        Same as `output_file` but for ``sys.stderr``, and `None` is not accepted.

    :param raw_output:

        If `True`, results are written to the output file raw, without adding
        whitespaces or newlines between yielded strings. Default is `False`.

    :param completion:

        If `True`, shell tab completion is enabled. Default is `True`. (You
        will also need to install it.)  See :mod:`argh.completion`.

    :param skip_unknown_args:

        If `True`, unknown arguments do not cause an error
        (`ArgumentParser.parse_known_args` is used).

    :param namespace:

        .. deprecated:: 0.31

          This argument will be removed soon after v0.31.

    :param always_flush:

        If the output stream is not a terminal (i.e. redirected to a file or
        another process), it's probably buffered.  In most cases it doesn't
        matter.

        However, if the output of your program is generated with delays
        between the lines and you may want to redirect them to another process
        and immediately see the results (e.g. `my_app.py | grep something`),
        it's a good idea to force flushing of the buffer.

        .. versionadded:: 0.31

    By default the exceptions are not wrapped and will propagate. The only
    exception that is always wrapped is :class:`~argh.exceptions.CommandError`
    which is interpreted as an expected event so the traceback is hidden.
    You can also mark arbitrary exceptions as "wrappable" by using the
    :func:`~argh.decorators.wrap_errors` decorator.

    Wrapped exceptions, or other "expected errors" like parse failures,
    will cause a SystemExit to be raised.
    """
    if namespace:
        warnings.warn(
            DeprecationWarning(
                "The argument `namespace` in `dispatch()` is deprecated. "
                "It will be removed in the next minor version after v0.31."
            )
        )

    # TODO: remove in v0.31+/v1.0
    if add_help_command:  # pragma: nocover
        warnings.warn(
            DeprecationWarning(
                "The argument `add_help_command` in `dispatch()` is deprecated. "
                "The user is expected to type `--help` instead of `help`."
            )
        )
        if argv and argv[0] == "help":
            argv.pop(0)
            argv.append("--help")

    endpoint_function, namespace_obj = parse_and_resolve(
        parser=parser,
        completion=completion,
        argv=argv,
        namespace=namespace,
        skip_unknown_args=skip_unknown_args,
    )

    if not endpoint_function:
        parser.print_usage(output_file)
        return None

    return run_endpoint_function(
        function=endpoint_function,
        namespace_obj=namespace_obj,
        output_file=output_file,
        errors_file=errors_file,
        raw_output=raw_output,
        always_flush=always_flush,
    )


def parse_and_resolve(
    parser: argparse.ArgumentParser,
    argv: Optional[List[str]] = None,
    completion: bool = True,
    namespace: Optional[argparse.Namespace] = None,
    skip_unknown_args: bool = False,
) -> Tuple[Optional[Callable], argparse.Namespace]:
    """
    .. versionadded:: 0.30

    Parses CLI arguments and resolves the endpoint function.

    :param namespace:

        .. deprecated:: 0.31

          This argument will be removed soon after v0.31.
    """
    if completion:
        autocomplete(parser)

    if argv is None:
        argv = sys.argv[1:]

    if not namespace:
        namespace = ArghNamespace()

    # this will raise SystemExit if parsing fails
    if skip_unknown_args:
        namespace_obj, unknown_args = parser.parse_known_args(argv, namespace=namespace)
        # store unknown args on the namespace
        namespace_obj._unknown_args = unknown_args
    else:
        namespace_obj = parser.parse_args(argv, namespace=namespace)

    function = _get_function_from_namespace_obj(namespace_obj)

    return function, namespace_obj


def run_endpoint_function(
    function: Callable,
    namespace_obj: argparse.Namespace,
    output_file: IO = sys.stdout,
    errors_file: IO = sys.stderr,
    raw_output: bool = False,
    always_flush: bool = False,
) -> Optional[str]:
    """
    .. versionadded:: 0.30

    Extracts arguments from the namespace object, calls the endpoint function
    and processes its output.

    :param always_flush:

        Flush the buffer after every line even if `output_file` is not a TTY.
        Turn this off if you don't need dynamic output)
    """
    lines = _execute_command(function, namespace_obj, errors_file)

    return _process_command_output(lines, output_file, raw_output, always_flush)


def _process_command_output(
    lines: Iterator[str],
    output_file: Optional[IO],
    raw_output: bool,
    always_flush: bool,
) -> Optional[str]:
    out_io: IO

    if output_file is None:
        # user wants a string; we create an internal temporary file-like object
        # and will return its contents as a string
        out_io = io.StringIO()
    else:
        # normally this is stdout; can be any file
        out_io = output_file

    # this may raise user exceptions, or SystemExit for wrapped exceptions
    for line in lines:
        # print the line as soon as it is generated to ensure that it is
        # displayed to the user before anything else happens, e.g.
        # raw_input() is called
        out_io.write(str(line))
        if not raw_output:
            # in most cases user wants one message per line
            out_io.write("\n")

        # If it's not a terminal (i.e. redirected to a file or another
        # process), it's probably buffered.  In most cases it doesn't matter
        # but if the output is generated with delays between the lines and we
        # may want to monitor it (e.g. `my_app.py | grep something`), it's a
        # good idea to force flushing.
        if always_flush:
            out_io.flush()

    if output_file is None:
        # user wanted a string; return contents of our temporary file-like obj
        out_io.seek(0)
        return out_io.read()

    return None


def _get_function_from_namespace_obj(
    namespace_obj: argparse.Namespace,
) -> Optional[Callable]:
    if isinstance(namespace_obj, ArghNamespace):
        # our special namespace object keeps the stack of assigned functions
        try:
            function = namespace_obj.get_function()
        except (AttributeError, IndexError):
            return None
    else:
        # a custom (probably vanilla) namespace object keeps the last assigned
        # function; this may be wrong but at least something may work
        if not hasattr(namespace_obj, DEST_FUNCTION):
            return None
        function = getattr(namespace_obj, DEST_FUNCTION)

    if not hasattr(function, "__call__"):
        return None

    return function


def _execute_command(
    function: Callable, namespace_obj: argparse.Namespace, errors_file: IO
) -> Iterator[str]:
    """
    Assumes that `function` is a callable.  Tries different approaches
    to call it (with `namespace_obj` or with ordinary signature).
    Yields the results line by line.

    If :class:`~argh.exceptions.CommandError` is raised, its message is
    written to the error file, and a SystemExit is raised.
    All other exceptions propagate unless marked as wrappable
    by :func:`wrap_errors`.
    """

    # the function is nested to catch certain exceptions (see below)
    def _call():
        # Actually call the function

        # namespace -> dictionary
        def _flat_key(key):
            return key.replace("-", "_")

        values_by_arg_name = dict(
            (_flat_key(k), v) for k, v in vars(namespace_obj).items()
        )

        # filter the namespace variables so that only those expected
        # by the actual function will pass

        func_signature = inspect.signature(function)
        func_params = func_signature.parameters.values()

        positional_names = [
            p.name
            for p in func_params
            if p.kind in (p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD)
        ]
        kwonly_names = [p.name for p in func_params if p.kind == p.KEYWORD_ONLY]
        varargs_names = [p.name for p in func_params if p.kind == p.VAR_POSITIONAL]
        positional_values = [values_by_arg_name[name] for name in positional_names]
        values_by_name = dict((k, values_by_arg_name[k]) for k in kwonly_names)

        # *args
        if varargs_names:
            value = varargs_names[0]
            positional_values += values_by_arg_name[value]

        # **kwargs
        if any(p for p in func_params if p.kind == p.VAR_KEYWORD):
            not_kwargs = (
                [DEST_FUNCTION] + positional_names + varargs_names + kwonly_names
            )
            for k in vars(namespace_obj):
                normalized_k = _flat_key(k)
                if k.startswith("_") or normalized_k in not_kwargs:
                    continue
                values_by_name[normalized_k] = getattr(namespace_obj, k)

        result = function(*positional_values, **values_by_name)

        # Yield the results
        if isinstance(result, (GeneratorType, list, tuple)):
            # yield each line ASAP, convert CommandError message to a line
            for line in result:
                yield line
        else:
            # yield non-empty non-iterable result as a single line
            if result is not None:
                yield result

    wrappable_exceptions = [CommandError]
    wrappable_exceptions += getattr(function, ATTR_WRAPPED_EXCEPTIONS, [])

    def default_exception_processor(exc: Exception) -> str:
        return f"{exc.__class__.__name__}: {exc}"

    try:
        result = _call()
        for line in result:
            yield line
    except tuple(wrappable_exceptions) as exc:
        processor = getattr(
            function, ATTR_WRAPPED_EXCEPTIONS_PROCESSOR, default_exception_processor
        )

        errors_file.write(str(processor(exc)))
        errors_file.write("\n")

        # Use code from CommandError if available, otherwise default to 1
        code = exc.code if isinstance(exc, CommandError) and exc.code is not None else 1

        sys.exit(code)


def dispatch_command(
    function: Callable, *args, old_name_mapping_policy=True, **kwargs
) -> None:
    """
    A wrapper for :func:`dispatch` that creates a one-command parser.
    Uses :attr:`argh.constants.PARSER_FORMATTER`.

    :param old_name_mapping_policy:

        .. versionadded:: 0.31

        If `True`, sets the default argument naming policy to
        `~argh.assembling.NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT`, otherwise
        to `~argh.assembling.NameMappingPolicy.BY_NAME_IF_KWONLY`.

        .. warning:: tho default will be changed to `False` in v.0.33 (or v.1.0).

    This::

        dispatch_command(foo)

    ...is a shortcut for::

        parser = ArgumentParser()
        set_default_command(parser, foo, name_mapping_policy=...)
        dispatch(parser)

    This function can be also used as a decorator::

        @dispatch_command
        def main(foo: int = 123) -> int:
            return foo + 1

    """
    if old_name_mapping_policy:
        name_mapping_policy = NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT
    else:
        name_mapping_policy = NameMappingPolicy.BY_NAME_IF_KWONLY

    parser = argparse.ArgumentParser(formatter_class=PARSER_FORMATTER)
    set_default_command(parser, function, name_mapping_policy=name_mapping_policy)
    dispatch(parser, *args, **kwargs)


def dispatch_commands(
    functions: List[Callable], *args, old_name_mapping_policy=True, **kwargs
) -> None:
    """
    A wrapper for :func:`dispatch` that creates a parser, adds commands to
    the parser and dispatches them.
    Uses :attr:`PARSER_FORMATTER`.

    :param old_name_mapping_policy:

        .. versionadded:: 0.31

        If `True`, sets the default argument naming policy to
        `~argh.assembling.NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT`, otherwise
        to `~argh.assembling.NameMappingPolicy.BY_NAME_IF_KWONLY`.

        .. warning:: tho default will be changed to `False` in v.0.33 (or v.1.0).

    This::

        dispatch_commands([foo, bar])

    ...is a shortcut for::

        parser = ArgumentParser()
        add_commands(parser, [foo, bar], name_mapping_policy=...)
        dispatch(parser)

    """
    if old_name_mapping_policy:
        name_mapping_policy = NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT
    else:
        name_mapping_policy = NameMappingPolicy.BY_NAME_IF_KWONLY

    parser = argparse.ArgumentParser(formatter_class=PARSER_FORMATTER)
    add_commands(parser, functions, name_mapping_policy=name_mapping_policy)
    dispatch(parser, *args, **kwargs)


class EntryPoint:
    """
    An object to which functions can be attached and then dispatched.

    When called with an argument, the argument (a function) is registered
    at this entry point as a command.

    When called without an argument, dispatching is triggered with all
    previously registered commands.

    Usage::

        from argh import EntryPoint

        app = EntryPoint("main", {"description": "This is a cool app"})

        @app
        def ls() -> Iterator[int]:
            for i in range(10):
                yield i

        @app
        def greet() -> str:
            return "hello"

        if __name__ == "__main__":
            app()

    """

    def __init__(
        self, name: Optional[str] = None, parser_kwargs: Optional[Dict[str, Any]] = None
    ) -> None:
        self.name = name or "unnamed"
        self.commands: List[Callable] = []
        self.parser_kwargs = parser_kwargs or {}

    def __call__(self, function: Optional[Callable] = None):
        if function:
            self._register_command(function)
            return function

        return self._dispatch()

    def _register_command(self, function: Callable) -> None:
        self.commands.append(function)

    def _dispatch(self) -> None:
        if not self.commands:
            raise DispatchingError(f'no commands for entry point "{self.name}"')

        parser = argparse.ArgumentParser(**self.parser_kwargs)
        add_commands(parser, self.commands)
        dispatch(parser)



"""
Data transfer objects for internal usage.
"""

from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Type, Union


class NotDefined:
    """
    Specifies that an argument should not be passed to
    ArgumentParser.add_argument(), even as None
    """


@dataclass
class ParserAddArgumentSpec:
    """
    DTO, maps CLI arg(s) onto a function arg.
    Ends up in ArgumentParser.add_argument().
    """

    func_arg_name: Optional[str]  # TODO: make it required (needs rearranging the logic)
    cli_arg_names: List[str]
    is_required: Union[bool, Type[NotDefined]] = NotDefined
    default_value: Any = NotDefined
    nargs: Optional[str] = None
    other_add_parser_kwargs: Dict[str, Any] = field(default_factory=dict)

    # https://kislyuk.github.io/argcomplete/#specifying-completers
    completer: Optional[Callable] = None

    def update(self, other: "ParserAddArgumentSpec") -> None:
        for name in other.cli_arg_names:
            if name not in self.cli_arg_names:
                self.cli_arg_names.append(name)

        if other.is_required != NotDefined:
            self.is_required = other.is_required

        if other.default_value != NotDefined:
            self.default_value = other.default_value

        if other.nargs:
            self.nargs = other.nargs

        if other.completer:
            self.completer = other.completer

        self.other_add_parser_kwargs.update(other.other_add_parser_kwargs)

    def get_all_kwargs(self) -> Dict[str, Any]:
        kwargs: Dict[str, Any] = {}

        if self.is_required != NotDefined:
            kwargs["required"] = self.is_required

        if self.default_value != NotDefined:
            kwargs["default"] = self.default_value

        if self.nargs:
            kwargs["nargs"] = self.nargs

        return dict(kwargs, **self.other_add_parser_kwargs)

    @classmethod
    def make_from_kwargs(
        cls, func_arg_name, cli_arg_names, parser_add_argument_kwargs: Dict[str, Any]
    ) -> "ParserAddArgumentSpec":
        """
        Constructs and returns a `ParserAddArgumentSpec` instance
        according to keyword arguments according to the
        `ArgumentParser.add_argument()` signature.
        """
        kwargs_copy = parser_add_argument_kwargs.copy()
        instance = cls(
            func_arg_name=func_arg_name,
            cli_arg_names=cli_arg_names,
        )
        if "required" in kwargs_copy:
            instance.is_required = kwargs_copy.pop("required")
        if "nargs" in kwargs_copy:
            instance.nargs = kwargs_copy.pop("nargs")
        if "default" in kwargs_copy:
            instance.default_value = kwargs_copy.pop("default")
        if kwargs_copy:
            instance.other_add_parser_kwargs = kwargs_copy
        return instance



#
#  Copyright © 2010—2023 Andrey Mikhaylenko and contributors
#
#  This file is part of Argh.
#
#  Argh is free software under terms of the GNU Lesser
#  General Public License version 3 (LGPLv3) as published by the Free
#  Software Foundation. See the file README.rst for copying conditions.
#
"""
Assembling
~~~~~~~~~~

Functions and classes to properly assemble your commands in a parser.
"""

import inspect
import textwrap
import warnings
from argparse import OPTIONAL, ZERO_OR_MORE, ArgumentParser
from collections import OrderedDict
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
    get_args,
    get_origin,
)

# types.UnionType was introduced in Python < 3.10
try:  # pragma: no cover
    from types import UnionType

    UNION_TYPES = [Union, UnionType]
except ImportError:  # pragma: no cover
    UNION_TYPES = [Union]

from argh.completion import COMPLETION_ENABLED
from argh.constants import (
    ATTR_ALIASES,
    ATTR_ARGS,
    ATTR_NAME,
    DEFAULT_ARGUMENT_TEMPLATE,
    DEST_FUNCTION,
    PARSER_FORMATTER,
)
from argh.dto import NotDefined, ParserAddArgumentSpec
from argh.exceptions import AssemblingError
from argh.utils import get_subparsers

__all__ = [
    "set_default_command",
    "add_commands",
    "add_subcommands",
    "NameMappingPolicy",
]


class NameMappingPolicy(Enum):
    """
    Represents possible approaches to treat default values when inferring argument
    specification from function signature.

    * `BY_NAME_IF_KWONLY` is the default and recommended approach introduced
      in v0.30.  It enables fine control over two aspects:

      * positional vs named;
      * required vs optional.

      "Normal" arguments are identified by position, "kwonly" are identified by
      name, regardless of the presence of default values.  A positional with a
      default value becomes optional but still positional (``nargs=OPTIONAL``).
      A kwonly argument without a default value becomes a required named
      argument.

      Example::

          def func(alpha, beta=1, *, gamma, delta=2): ...

      is equivalent to::

          prog alpha [beta] --gamma [--delta DELTA]

      That is, ``alpha`` and ``--gamma`` are mandatory while ``beta`` and
      ``--delta`` are optional (they have default values).

    * `BY_NAME_IF_HAS_DEFAULT` is very close to the the legacy approach
      (pre-v0.30).  If a function argument has a default value, it becomes an
      "option" (called by name, like ``--foo``); otherwise it's treated as a
      positional argument.

      Example::

          def func(alpha, beta=1, *, gamma, delta=2): ...

      is equivalent to::

          prog [--beta BETA] [--delta DELTA] alpha gamma

      That is, ``alpha`` and ``gamma`` are mandatory and positional, while
      ``--beta`` and ``--delta`` are optional (they have default values).  Note
      that it's impossible to have an optional positional or a mandatory named
      argument.

      The difference between this policy and the behaviour of Argh before
      v0.30 is in the treatment of kwonly arguments without default values:
      they used to become ``--foo FOO`` (required) but for the sake of
      simplicity they are treated as positionals.  If you are already using
      kwonly args, please consider the better suited policy `BY_NAME_IF_KWONLY`
      instead.

    It is recommended to migrate any older code to `BY_NAME_IF_KWONLY`.

    .. versionadded:: 0.30
    """

    BY_NAME_IF_HAS_DEFAULT = "specify CLI argument by name if it has a default value"
    BY_NAME_IF_KWONLY = "specify CLI argument by name if it comes from kwonly"


def infer_argspecs_from_function(
    function: Callable,
    name_mapping_policy: Optional[NameMappingPolicy] = None,
    can_use_hints: bool = False,
) -> Iterator[ParserAddArgumentSpec]:
    if name_mapping_policy and name_mapping_policy not in NameMappingPolicy:
        raise NotImplementedError(f"Unknown name mapping policy {name_mapping_policy}")

    func_signature = inspect.signature(function)
    has_kwonly = any(
        p.kind == p.KEYWORD_ONLY for p in func_signature.parameters.values()
    )

    # define the list of conflicting option strings
    # (short forms, i.e. single-character ones)
    named_args = [
        p.name
        for p in func_signature.parameters.values()
        if p.default is not p.empty or p.kind == p.KEYWORD_ONLY
    ]
    named_arg_chars = [a[0] for a in named_args]
    named_arg_char_counts = dict(
        (char, named_arg_chars.count(char)) for char in set(named_arg_chars)
    )
    conflicting_opts = tuple(
        char for char in named_arg_char_counts if 1 < named_arg_char_counts[char]
    )

    def _make_cli_arg_names_options(arg_name) -> Tuple[List[str], List[str]]:
        cliified_arg_name = arg_name.replace("_", "-")
        positionals = [cliified_arg_name]
        can_have_short_opt = arg_name[0] not in conflicting_opts

        if can_have_short_opt:
            options = [f"-{cliified_arg_name[0]}", f"--{cliified_arg_name}"]
        else:
            options = [f"--{cliified_arg_name}"]

        return positionals, options

    default_value: Any
    for parameter in func_signature.parameters.values():
        (cli_arg_names_positional, cli_arg_names_options) = _make_cli_arg_names_options(
            parameter.name
        )
        if parameter.default is not parameter.empty:
            default_value = parameter.default
        else:
            default_value = NotDefined

        extra_spec_kwargs = {}

        if can_use_hints:
            hints = function.__annotations__
            if parameter.name in hints:
                extra_spec_kwargs = (
                    TypingHintArgSpecGuesser.typing_hint_to_arg_spec_params(
                        hints[parameter.name]
                    )
                )

        if parameter.kind in (
            parameter.POSITIONAL_ONLY,
            parameter.POSITIONAL_OR_KEYWORD,
        ):
            if default_value != NotDefined and not name_mapping_policy:
                message = textwrap.dedent(
                    f"""
                    Argument "{parameter.name}" in function "{function.__name__}"
                    is not keyword-only but has a default value.

                    Please note that since Argh v.0.30 the default name mapping
                    policy has changed.

                    More information:
                    https://argh.readthedocs.io/en/latest/changes.html#version-0-30-0-2023-10-21

                    You need to upgrade your functions so that the arguments
                    that have default values become keyword-only:

                        f(x=1) -> f(*, x=1)

                    If you actually want an optional positional argument,
                    please set the name mapping policy explicitly to `BY_NAME_IF_KWONLY`.

                    If you choose to postpone the migration, you have two options:

                    a) set the policy explicitly to `BY_NAME_IF_HAS_DEFAULT`;
                    b) pin Argh version to 0.29 until you are ready to migrate.

                    Thank you for understanding!
                    """
                ).strip()

                # Assume legacy policy and show a warning if the signature is
                # simple (without kwonly args) so that the script continues working
                # but the author is urged to upgrade it.
                # When it cannot be auto-resolved (kwonly args mixed with old-style
                # ones but no policy specified), throw an error.
                #
                # TODO: remove in v.0.33 if it happens, otherwise in v1.0.
                if has_kwonly:
                    raise ArgumentNameMappingError(message)
                warnings.warn(DeprecationWarning(message))
                name_mapping_policy = NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT

            arg_spec = ParserAddArgumentSpec(
                func_arg_name=parameter.name,
                cli_arg_names=cli_arg_names_positional,
                default_value=default_value,
                other_add_parser_kwargs=extra_spec_kwargs,
            )

            if default_value != NotDefined:
                if name_mapping_policy == NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT:
                    arg_spec.cli_arg_names = cli_arg_names_options
                else:
                    arg_spec.nargs = OPTIONAL

            # annotations are interpreted without regard to the broader
            # context, e.g. default values; in some cases argparse requires
            # pretty specific combinations of props, so we need to adjust them
            if can_use_hints:
                # "required" is invalid for positional CLI argument;
                # it may have been set from Optional[...] hint above.
                # Reinterpret it as "optional positional" instead.
                if "required" in arg_spec.other_add_parser_kwargs:
                    value = arg_spec.other_add_parser_kwargs.pop("required")
                    if value is False:
                        arg_spec.nargs = OPTIONAL

                if name_mapping_policy == NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT:
                    # The guesser yields `type=bool` from `foo: bool = False`
                    # but `type` is incompatible with `action="store_true"` which
                    # is added by guess_extra_parser_add_argument_spec_kwargs().
                    if (
                        isinstance(arg_spec.default_value, bool)
                        and arg_spec.other_add_parser_kwargs.get("type") == bool
                    ):
                        del arg_spec.other_add_parser_kwargs["type"]

            yield arg_spec

        elif parameter.kind == parameter.KEYWORD_ONLY:
            arg_spec = ParserAddArgumentSpec(
                func_arg_name=parameter.name,
                cli_arg_names=cli_arg_names_positional,
                default_value=default_value,
                other_add_parser_kwargs=extra_spec_kwargs,
            )

            if name_mapping_policy == NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT:
                if default_value != NotDefined:
                    arg_spec.cli_arg_names = cli_arg_names_options
            else:
                arg_spec.cli_arg_names = cli_arg_names_options
                if default_value == NotDefined:
                    arg_spec.is_required = True

            # annotations are interpreted without regard to the broader
            # context, e.g. default values; in some cases argparse requires
            # pretty specific combinations of props, so we need to adjust them
            if can_use_hints:
                # The guesser yields `type=bool` from `foo: bool = False`
                # but `type` is incompatible with `action="store_true"` which
                # is added by guess_extra_parser_add_argument_spec_kwargs().
                if (
                    isinstance(arg_spec.default_value, bool)
                    and arg_spec.other_add_parser_kwargs.get("type") == bool
                ):
                    del arg_spec.other_add_parser_kwargs["type"]

            yield arg_spec

        elif parameter.kind == parameter.VAR_POSITIONAL:
            yield ParserAddArgumentSpec(
                func_arg_name=parameter.name,
                cli_arg_names=[parameter.name.replace("_", "-")],
                nargs=ZERO_OR_MORE,
                other_add_parser_kwargs=extra_spec_kwargs,
            )


def guess_extra_parser_add_argument_spec_kwargs(
    parser_add_argument_spec: ParserAddArgumentSpec,
) -> Dict[str, Any]:
    """
    Given an argument specification, returns types, actions, etc. that could be
    guessed from it:

    * ``default=3`` → ``type=int``

      TODO: deprecate in favour of ``foo: int = 3`` in func signature.

    * ``choices=[3]`` → ``type=int``

      TODO: deprecate in favour of ``foo: int`` in func signature.

    * ``type=bool`` → ``action="store_false"`` or ``action="store_true"``
      (if action was not explicitly defined).

    """
    # TODO: use typing to extract
    other_add_parser_kwargs = parser_add_argument_spec.other_add_parser_kwargs
    guessed: Dict[str, Any] = {}
    is_positional = not parser_add_argument_spec.cli_arg_names[0].startswith("-")

    # Parser actions that accept argument 'type'
    TYPE_AWARE_ACTIONS = "store", "append"

    # guess type/action from default value
    default_value = parser_add_argument_spec.default_value
    if default_value not in (None, NotDefined):
        if isinstance(default_value, bool):
            if not is_positional and other_add_parser_kwargs.get("action") is None:
                # infer action from default value
                # (not applicable to positionals: _StoreAction doesn't accept `nargs`)
                guessed["action"] = "store_false" if default_value else "store_true"
        elif other_add_parser_kwargs.get("type") is None:
            if isinstance(default_value, (list, tuple)):
                if "nargs" not in other_add_parser_kwargs:
                    # the argument has a default value, so it doesn't have to
                    # be passed; "zero or more" is a reasonable guess
                    guessed["nargs"] = ZERO_OR_MORE
            else:
                # infer type from default value
                # (make sure that action handler supports this keyword)
                if other_add_parser_kwargs.get("action", "store") in TYPE_AWARE_ACTIONS:
                    guessed["type"] = type(default_value)

    # guess type from choices (first item)
    if other_add_parser_kwargs.get("choices") and "type" not in list(guessed) + list(
        other_add_parser_kwargs
    ):
        guessed["type"] = type(other_add_parser_kwargs["choices"][0])

    return guessed


def set_default_command(
    parser,
    function: Callable,
    name_mapping_policy: Optional[NameMappingPolicy] = None,
) -> None:
    """
    Sets default command (i.e. a function) for given parser.

    If `parser.description` is empty and the function has a docstring,
    it is used as the description.

    :param function:

        The function to use as the command.

    :name_mapping_policy:

        The policy to use when mapping function arguments onto CLI arguments.
        See :class:`.NameMappingPolicy`.  If not defined explicitly,
        `BY_NAME_IF_KWONLY` is used.

        .. versionadded:: 0.30

        .. versionchanged:: 0.30.2
           Raises `ArgumentNameMappingError` if the policy was not explicitly
           defined and a non-kwonly argument has a default value.  The reason
           is that it's very likely to be a case of non-migrated code where
           the argument was intended to be mapped onto a CLI option.  It's
           better to fail explicitly than to silently change the CLI API.

    .. note::

       If there are both explicitly declared arguments (e.g. via
       :func:`~argh.decorators.arg`) and ones inferred from the function
       signature, declared ones will be merged into inferred ones.
       If an argument does not conform to the function signature,
       `ArgumentNameMappingError` is raised.

    .. note::

       If the parser was created with ``add_help=True`` (which is by default),
       option name ``-h`` is silently removed from any argument.

    """
    func_signature = inspect.signature(function)

    # the **kwargs thing
    has_varkw = any(p.kind == p.VAR_KEYWORD for p in func_signature.parameters.values())

    declared_args: List[ParserAddArgumentSpec] = getattr(function, ATTR_ARGS, [])

    # transitional period: hints are used for types etc. only if @arg is not used
    can_use_hints = not declared_args

    inferred_args: List[ParserAddArgumentSpec] = list(
        infer_argspecs_from_function(
            function,
            name_mapping_policy=name_mapping_policy,
            can_use_hints=can_use_hints,
        )
    )

    if declared_args and not inferred_args and not has_varkw:
        raise ArgumentNameMappingError(
            f"{function.__name__}: cannot extend argument declarations "
            "for an endpoint function that takes no arguments."
        )

    if not declared_args:
        parser_add_argument_specs = inferred_args
    else:
        # We've got a mixture of declared and inferred arguments
        try:
            parser_add_argument_specs = _merge_inferred_and_declared_args(
                inferred_args=inferred_args,
                declared_args=declared_args,
                has_varkw=has_varkw,
            )
        except ArgumentNameMappingError as exc:
            raise ArgumentNameMappingError(f"{function.__name__}: {exc}") from exc

    # add the fully formed argument specs to the parser
    for spec in parser_add_argument_specs:
        _extend_parser_add_argument_spec(
            spec=spec, parser_adds_help_arg=parser.add_help
        )

        try:
            action = parser.add_argument(
                *spec.cli_arg_names,
                **spec.get_all_kwargs(),
            )
        except Exception as exc:
            err_cli_args = "/".join(spec.cli_arg_names)
            raise AssemblingError(
                f"{function.__name__}: cannot add '{spec.func_arg_name}' as {err_cli_args}: {exc}"
            ) from exc

        if COMPLETION_ENABLED and spec.completer:
            action.completer = spec.completer

    # display endpoint function docstring in command help
    docstring = inspect.getdoc(function)
    if docstring and not parser.description:
        parser.description = docstring

    # add the endpoint function to the parsing result (namespace)
    parser.set_defaults(
        **{
            DEST_FUNCTION: function,
        }
    )


def _extend_parser_add_argument_spec(
    spec: ParserAddArgumentSpec, parser_adds_help_arg: bool
) -> None:
    # add types, actions, etc. (e.g. default=3 implies type=int)
    spec.other_add_parser_kwargs.update(
        guess_extra_parser_add_argument_spec_kwargs(spec)
    )

    # display default value for this argument in command help
    if "help" not in spec.other_add_parser_kwargs:
        spec.other_add_parser_kwargs.update(help=DEFAULT_ARGUMENT_TEMPLATE)

    # If the parser was created with `add_help=True`, it automatically adds
    # the -h/--help argument (on argparse side).  If we have added -h for
    # another argument (e.g. --host) earlier (inferred or declared), we
    # need to remove that short form now.
    if parser_adds_help_arg and "-h" in spec.cli_arg_names:
        spec.cli_arg_names = [name for name in spec.cli_arg_names if name != "-h"]


def _merge_inferred_and_declared_args(
    inferred_args: List[ParserAddArgumentSpec],
    declared_args: List[ParserAddArgumentSpec],
    has_varkw: bool,
) -> List[ParserAddArgumentSpec]:
    # a mapping of "dest" strings to argument declarations.
    #
    # * a "dest" string is a normalized form of argument name, i.e.:
    #
    #     "-f", "--foo" → "foo"
    #     "foo-bar"     → "foo_bar"
    #
    # * argument declaration is a dictionary representing an argument;
    #   it is obtained either from infer_argspecs_from_function()
    #   or from an @arg decorator (as is).
    #
    specs_by_func_arg_name = OrderedDict()

    # arguments inferred from function signature
    for parser_add_argument_spec in inferred_args:
        specs_by_func_arg_name[parser_add_argument_spec.func_arg_name] = (
            parser_add_argument_spec
        )

    # arguments declared via @arg decorator
    for declared_spec in declared_args:
        parser_add_argument_spec = declared_spec
        func_arg_name = parser_add_argument_spec.func_arg_name

        if func_arg_name in specs_by_func_arg_name:
            # the argument is already known from function signature
            #
            # now make sure that this declared arg conforms to the function
            # signature and therefore only refines an inferred arg:
            #
            #      @arg("my-foo")    maps to  func(my_foo)
            #      @arg("--my-bar")  maps to  func(my_bar=...)

            # either both arguments are positional or both are optional
            decl_positional = _is_positional(declared_spec.cli_arg_names)
            infr_positional = _is_positional(
                specs_by_func_arg_name[func_arg_name].cli_arg_names
            )
            if decl_positional != infr_positional:
                kinds = {True: "positional", False: "optional"}
                kind_inferred = kinds[infr_positional]
                kind_declared = kinds[decl_positional]
                raise ArgumentNameMappingError(
                    f'argument "{func_arg_name}" declared as {kind_inferred} '
                    f"(in function signature) and {kind_declared} (via decorator). "
                    "If you've just migrated from Argh v.0.29, please check "
                    "the new default NameMappingPolicy. Perhaps you need "
                    "to replace `func(x=1)` with `func(*, x=1)`?"
                )

            # merge explicit argument declaration into the inferred one
            # (e.g. `help=...`)
            specs_by_func_arg_name[func_arg_name].update(parser_add_argument_spec)
        else:
            # the argument is not in function signature
            if has_varkw:
                # function accepts **kwargs; the argument goes into it
                specs_by_func_arg_name[func_arg_name] = parser_add_argument_spec
            else:
                # there's no way we can map the argument declaration
                # to function signature
                dest_option_strings = (
                    specs_by_func_arg_name[x].cli_arg_names
                    for x in specs_by_func_arg_name
                )
                msg_flags = ", ".join(declared_spec.cli_arg_names)
                msg_signature = ", ".join("/".join(x) for x in dest_option_strings)
                raise ArgumentNameMappingError(
                    f"argument {msg_flags} does not fit "
                    f"function signature: {msg_signature}"
                )

    # pack the modified data back into a list
    return list(specs_by_func_arg_name.values())


def _is_positional(args: List[str], prefix_chars: str = "-") -> bool:
    if not args or not args[0]:
        raise ValueError("Expected at least one argument")
    if args[0][0].startswith(tuple(prefix_chars)):
        return False
    return True


def add_commands(
    parser: ArgumentParser,
    functions: List[Callable],
    name_mapping_policy: Optional[NameMappingPolicy] = None,
    group_name: Optional[str] = None,
    group_kwargs: Optional[Dict[str, Any]] = None,
    func_kwargs: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Adds given functions as commands to given parser.

    :param parser:

        an :class:`argparse.ArgumentParser` instance.

    :param functions:

        a list of functions. A subparser is created for each of them.
        If the function is decorated with :func:`~argh.decorators.arg`, the
        arguments are passed to :meth:`argparse.ArgumentParser.add_argument`.
        See also :func:`~argh.dispatching.dispatch` for requirements
        concerning function signatures. The command name is inferred from the
        function name. Note that the underscores in the name are replaced with
        hyphens, i.e. function name "foo_bar" becomes command name "foo-bar".

    :param name_mapping_policy:

        See :class:`argh.assembling.NameMappingPolicy`.

        .. versionadded:: 0.30

    :param group_name:

        an optional string representing the group of commands. For example, if
        a command named "hello" is added without the group name, it will be
        available as "prog.py hello"; if the group name if specified as "greet",
        then the command will be accessible as "prog.py greet hello". The
        group itself is not callable, so "prog.py greet" will fail and only
        display a help message.

    :param func_kwargs:

        a `dict` of keyword arguments to be passed to each nested ArgumentParser
        instance created per command (i.e. per function).  Members of this
        dictionary have the highest priority, so a function's docstring is
        overridden by a `help` in `func_kwargs` (if present).

    :param group_kwargs:

        a `dict` of keyword arguments to be passed to the nested ArgumentParser
        instance under given `group_name`.

    .. note::

        This function modifies the parser object. Generally side effects are
        bad practice but we don't seem to have any choice as ArgumentParser is
        pretty opaque.
        You may prefer :meth:`~argh.helpers.ArghParser.add_commands` for a bit
        more predictable API.

    .. note::

       An attempt to add commands to a parser which already has a default
       function (e.g. added with :func:`~argh.assembling.set_default_command`)
       results in `AssemblingError`.

    """
    group_kwargs = group_kwargs or {}

    subparsers_action = get_subparsers(parser, create=True)

    if group_name:
        # Make a nested parser and init a deeper _SubParsersAction under it.

        # Create a named group of commands.  It will be listed along with
        # root-level commands in ``app.py --help``; in that context its `title`
        # can be used as a short description on the right side of its name.
        # Normally `title` is shown above the list of commands
        # in ``app.py my-group --help``.
        subsubparser = subparsers_action.add_parser(
            group_name, help=group_kwargs.get("title")
        )
        subparsers_action = subsubparser.add_subparsers(**group_kwargs)
    else:
        if group_kwargs:
            raise ValueError("`group_kwargs` only makes sense with `group_name`.")

    for func in functions:
        cmd_name, func_parser_kwargs = _extract_command_meta_from_func(func)

        # override any computed kwargs by manually supplied ones
        if func_kwargs:
            func_parser_kwargs.update(func_kwargs)

        # create and set up the parser for this command
        command_parser = subparsers_action.add_parser(cmd_name, **func_parser_kwargs)
        set_default_command(
            command_parser, func, name_mapping_policy=name_mapping_policy
        )


def _extract_command_meta_from_func(func: Callable) -> Tuple[str, dict]:
    # use explicitly defined name; if none, use function name (a_b → a-b)
    cmd_name = getattr(func, ATTR_NAME, func.__name__.replace("_", "-"))

    func_parser_kwargs: Dict[str, Any] = {
        # add command help from function's docstring
        "help": func.__doc__,
        # set default formatter
        "formatter_class": PARSER_FORMATTER,
    }

    # add aliases for command name
    func_parser_kwargs["aliases"] = getattr(func, ATTR_ALIASES, [])

    return cmd_name, func_parser_kwargs


def add_subcommands(
    parser: ArgumentParser, group_name: str, functions: List[Callable], **group_kwargs
) -> None:
    """
    A wrapper for :func:`add_commands`.

    These examples are equivalent::

        add_commands(
            parser,
            [get, put],
            group_name="db",
            group_kwargs={
                "title": "database commands",
                "help": "CRUD for our silly database"
            }
        )

        add_subcommands(
            parser,
            "db",
            [get, put],
            title="database commands",
            help="CRUD for our database"
        )

    """
    add_commands(parser, functions, group_name=group_name, group_kwargs=group_kwargs)


class ArgumentNameMappingError(AssemblingError): ...


class TypingHintArgSpecGuesser:
    BASIC_TYPES = (str, int, float, bool)

    @classmethod
    def typing_hint_to_arg_spec_params(
        cls, type_def: type, is_positional: bool = False
    ) -> Dict[str, Any]:
        origin = get_origin(type_def)
        args = get_args(type_def)

        # `str`
        if type_def in cls.BASIC_TYPES:
            return {
                "type": type_def
                # "type": _parse_basic_type(type_def)
            }

        # `list`
        if type_def in (list, List):
            return {"nargs": ZERO_OR_MORE}

        # `Literal["a", "b"]`
        if origin == Literal:
            return {"choices": args, "type": type(args[0])}

        # `str | int`
        if any(origin is t for t in UNION_TYPES):
            retval = {}
            first_subtype = args[0]
            if first_subtype in cls.BASIC_TYPES:
                retval["type"] = first_subtype

            if first_subtype in (list, List):
                retval["nargs"] = ZERO_OR_MORE

            if first_subtype != List and get_origin(first_subtype) == list:
                retval["nargs"] = ZERO_OR_MORE
                item_type = cls._extract_item_type_from_list_type(first_subtype)
                if item_type:
                    retval["type"] = item_type

            if type(None) in args:
                retval["required"] = False
            return retval

        # `list[str]`
        if origin == list:
            retval = {}
            retval["nargs"] = ZERO_OR_MORE
            if args[0] in cls.BASIC_TYPES:
                retval["type"] = args[0]
            return retval

        return {}

    @classmethod
    def _extract_item_type_from_list_type(cls, type_def) -> Optional[type]:
        args = get_args(type_def)
        if args[0] in cls.BASIC_TYPES:
            return args[0]
        return None
